
/* should be corrected
red		$ff0000
yellow	$ffff00 (red and green)
green	$00ff00
cyan	$00ffff (green and blue)
blue	$0000ff
magenta $ff00ff (red and blue)
*/

#include "Hue.h"

Hue::Hue(PClip _child, float _hue, float _sat, float _intensity, const char * _channels, IScriptEnvironment* env) :
	GenericVideoFilter(_child), hue(_hue), sat(_sat), intensity(_intensity), channels(_channels) {
	
	if (!vi.IsRGB()) {
		env->ThrowError("Your source is not RGB");
	}

	char ch = 1;

	doM = false;
	doB = false;
	doC = false;
	doG = false;
	doY = false;
	doR = false;
	
	for (int k=0; ch!='\0'; ++k) {
		ch = channels[k];
		if (ch == 'R')
			doR = true;
		if (ch == 'Y')
			doY = true;
		if (ch == 'G')
			doG = true;
		if (ch == 'C')
			doC = true;
		if (ch == 'B')
			doB = true;
		if (ch == 'M')
			doM = true;
	}

	if ((!doR)&&(!doY)&&(!doG)&&(!doC)&&(!doB)&&(!doM)) {
		env->ThrowError("Hue: channels must be \"R\", \"Y\", \"G\", \"C\", \"B\", \"M\" or a combination of it ");
	}
}

Hue::~Hue() {
}

PVideoFrame __stdcall Hue::GetFrame(int n, IScriptEnvironment* env) {

	PVideoFrame src = child->GetFrame(n, env);
		
	const unsigned char* srcp = src->GetReadPtr();
	const int src_pitch = src->GetPitch();
	const int src_width = src->GetRowSize();
	const int src_height = src->GetHeight();
	
	PVideoFrame dst = env->NewVideoFrame(vi);				
	unsigned char* dstp = dst->GetWritePtr();
	int dst_pitch = dst->GetPitch();
	int dst_width = dst->GetRowSize();
	int dst_height = dst->GetHeight();

	float mat[4][4];
	int lmat[4][4];

	int r, g, b, min;

	float s = sat;
	float i = intensity;

	bool doit;

	/* Create the transformation matrix. */
	identmat(mat);
	cscalemat(mat, i, i, i);
	saturatemat(mat, s);
	simplehuerotatemat(mat, hue);
	mat2lmat(mat, lmat);


	for (int h=0; h < src_height; h++) {     // Loop from bottom line to top line.
		for (int w=0; w < src_width; w+=4) {	// and from leftmost pixel to rightmost one. 

			b = *(srcp +w);
			g = *(srcp +w+1);
			r = *(srcp +w+2);

			min = r;
			doit = false;
			if (g < min) min = g;
			if (b < min) min = b;
			if (min == b)
			{
			/* In RGB, if all the values are the same, you have grey.
			So the min value is never the hue of the color. The
			top 2 create the hue. If blue is min, we don't worry
			if you selected a 'blue' check box, because the hue
			is not blue. Magenta, Blue and Cyan have blue in them,
			so with a min blue, they are ruled out. */
				if (g < 2*r/3)
				{
					if (doR) doit = true;
				}
				else if (r < 2*g/3)
				{
					if (doG) doit = true;
				}
				else if (doY) doit = true;
				/* Around the color wheel we get Red, Yellow, Green. Red + Green makes yellow.
				So it is broken up as 1/3, 1/3, 1/3. If g < 2/3 r then the hue of the pixel is 
				Red. If r < 2/3 g then the pixel is green. Else it is both, so it is yellow. */
			}
			else if (min == r)
			{
				if (g < 2*b/3)
				{
					if (doB) doit = true;
				}
				else if (b < 2*g/3)
				{
					if (doG) doit = true;
				}
				else if (doC) doit = true;
			}
			else
			{
				if (r < 2*b/3)
				{
					if (doB) doit = true;
				}
				else if (b < 2*r/3)
				{
					if (doR) doit = true;
				}
				else if (doM) doit = true;
			}

			if (doit == true)
				lxformRGB(lmat, &r, &g, &b);

			*(dstp +w)=b;
			*(dstp +w+1)=g;
			*(dstp +w+2)=r;

		}																
		srcp += src_pitch; // Add the pitch (note use of pitch and not width) of one line (in bytes) to the source pointer
		dstp += dst_pitch; // Add the pitch to the destination pointer.
	}

	return dst;
}


#define M_PI float (3.1415926)

#define RLUM float (0.3086)
#define GLUM float (0.6094)
#define BLUM float (0.0820)

/* 
 *	matrixmult -	
 *		multiply two matrices
 */
void matrixmult(float a[4][4], float b[4][4], float c[4][4])
{
    int x, y;
    float temp[4][4];

    for(y=0; y<4 ; y++)
        for(x=0 ; x<4 ; x++) {
            temp[y][x] = b[y][0] * a[0][x]
                       + b[y][1] * a[1][x]
                       + b[y][2] * a[2][x]
                       + b[y][3] * a[3][x];
        }
    for(y=0; y<4; y++)
        for(x=0; x<4; x++)
            c[y][x] = temp[y][x];
}

/* 
 *	identmat -	
 *		make an identity matrix
 */
void Hue::identmat(float mat[4][4])
{
	float *matrix = (float *) mat;

    *matrix++ = 1.0;    /* row 1        */
    *matrix++ = 0.0;
    *matrix++ = 0.0;
    *matrix++ = 0.0;
    *matrix++ = 0.0;    /* row 2        */
    *matrix++ = 1.0;
    *matrix++ = 0.0;
    *matrix++ = 0.0;
    *matrix++ = 0.0;    /* row 3        */
    *matrix++ = 0.0;
    *matrix++ = 1.0;
    *matrix++ = 0.0;
    *matrix++ = 0.0;    /* row 4        */
    *matrix++ = 0.0;
    *matrix++ = 0.0;
    *matrix++ = 1.0;
}

/* 
 *	cscalemat -	
 *		make a color scale marix
 */
void Hue::cscalemat(float mat[4][4], float rscale, float gscale, float bscale)
{
    float mmat[4][4];

    mmat[0][0] = rscale;
    mmat[0][1] = 0.0;
    mmat[0][2] = 0.0;
    mmat[0][3] = 0.0;

    mmat[1][0] = 0.0;
    mmat[1][1] = gscale;
    mmat[1][2] = 0.0;
    mmat[1][3] = 0.0;


    mmat[2][0] = 0.0;
    mmat[2][1] = 0.0;
    mmat[2][2] = bscale;
    mmat[2][3] = 0.0;

    mmat[3][0] = 0.0;
    mmat[3][1] = 0.0;
    mmat[3][2] = 0.0;
    mmat[3][3] = 1.0;
    matrixmult(mmat,mat,mat);
}

/* 
 *	saturatemat -	
 *		make a saturation marix
 */
void Hue::saturatemat(float mat[4][4], float s)
{
    float mmat[4][4];
    float a, b, c, d, e, f, g, h, i;
    float rwgt, gwgt, bwgt;

    rwgt = RLUM;
    gwgt = GLUM;
    bwgt = BLUM;

    a = (float) (1.0-s)*rwgt + s;
    b = (float) (1.0-s)*rwgt;
    c = (float) (1.0-s)*rwgt;
    d = (float) (1.0-s)*gwgt;
    e = (float) (1.0-s)*gwgt + s;
    f = (float) (1.0-s)*gwgt;
    g = (float) (1.0-s)*bwgt;
    h = (float) (1.0-s)*bwgt;
    i = (float) (1.0-s)*bwgt + s;
    mmat[0][0] = a;
    mmat[0][1] = b;
    mmat[0][2] = c;
    mmat[0][3] = 0.0;

    mmat[1][0] = d;
    mmat[1][1] = e;
    mmat[1][2] = f;
    mmat[1][3] = 0.0;

    mmat[2][0] = g;
    mmat[2][1] = h;
    mmat[2][2] = i;
    mmat[2][3] = 0.0;

    mmat[3][0] = 0.0;
    mmat[3][1] = 0.0;
    mmat[3][2] = 0.0;
    mmat[3][3] = 1.0;
    matrixmult(mmat,mat,mat);
}

/* 
 *	xrotate -	
 *		rotate about the x (red) axis
 */
void xrotatemat(float mat[4][4], float rs, float rc)
{
    float mmat[4][4];

    mmat[0][0] = 1.0;
    mmat[0][1] = 0.0;
    mmat[0][2] = 0.0;
    mmat[0][3] = 0.0;

    mmat[1][0] = 0.0;
    mmat[1][1] = rc;
    mmat[1][2] = rs;
    mmat[1][3] = 0.0;

    mmat[2][0] = 0.0;
    mmat[2][1] = -rs;
    mmat[2][2] = rc;
    mmat[2][3] = 0.0;

    mmat[3][0] = 0.0;
    mmat[3][1] = 0.0;
    mmat[3][2] = 0.0;
    mmat[3][3] = 1.0;
    matrixmult(mmat,mat,mat);
}

/* 
 *	yrotate -	
 *		rotate about the y (green) axis
 */
void yrotatemat(float mat[4][4], float rs, float rc)
{
    float mmat[4][4];

    mmat[0][0] = rc;
    mmat[0][1] = 0.0;
    mmat[0][2] = -rs;
    mmat[0][3] = 0.0;

    mmat[1][0] = 0.0;
    mmat[1][1] = 1.0;
    mmat[1][2] = 0.0;
    mmat[1][3] = 0.0;

    mmat[2][0] = rs;
    mmat[2][1] = 0.0;
    mmat[2][2] = rc;
    mmat[2][3] = 0.0;

    mmat[3][0] = 0.0;
    mmat[3][1] = 0.0;
    mmat[3][2] = 0.0;
    mmat[3][3] = 1.0;
    matrixmult(mmat,mat,mat);
}

/* 
 *	zrotate -	
 *		rotate about the z (blue) axis
 */
void zrotatemat(float mat[4][4], float rs, float rc)
{
    float mmat[4][4];

    mmat[0][0] = rc;
    mmat[0][1] = rs;
    mmat[0][2] = 0.0;
    mmat[0][3] = 0.0;

    mmat[1][0] = -rs;
    mmat[1][1] = rc;
    mmat[1][2] = 0.0;
    mmat[1][3] = 0.0;

    mmat[2][0] = 0.0;
    mmat[2][1] = 0.0;
    mmat[2][2] = 1.0;
    mmat[2][3] = 0.0;

    mmat[3][0] = 0.0;
    mmat[3][1] = 0.0;
    mmat[3][2] = 0.0;
    mmat[3][3] = 1.0;
    matrixmult(mmat,mat,mat);
}

/* 
 *	simplehuerotatemat -	
 *		simple hue rotation. This changes luminance 
 */
void Hue::simplehuerotatemat(float mat[4][4], float rot)
{
    float mag;
    float xrs, xrc;
    float yrs, yrc;
    float zrs, zrc;

/* rotate the grey vector into positive Z */
    mag = (float) sqrt(2.0);
    xrs = (float) 1.0/mag;
    xrc = (float) 1.0/mag;
    xrotatemat(mat,xrs,xrc);

    mag = (float) sqrt(3.0);
    yrs = (float) -1.0/mag;
    yrc = (float) sqrt(2.0)/mag;
    yrotatemat(mat,yrs,yrc);

/* rotate the hue */
    zrs = (float) sin(rot*M_PI/180.0);
    zrc = (float) cos(rot*M_PI/180.0);
    zrotatemat(mat,zrs,zrc);

/* rotate the grey vector back into place */
    yrotatemat(mat,-yrs,yrc);
    xrotatemat(mat,-xrs,xrc);
}

#define DENOM 0x10000 // 65536

inline void Hue::lxformRGB(int mat[4][4], int *r, int *g, int *b)
{
    int ir, ig, ib;

	ir = *r;
	ig = *g;
	ib = *b;
	*r = (int) (ir*mat[0][0] + ig*mat[1][0] + ib*mat[2][0])/DENOM;
	*g = (int) (ir*mat[0][1] + ig*mat[1][1] + ib*mat[2][1])/DENOM;
	*b = (int) (ir*mat[0][2] + ig*mat[1][2] + ib*mat[2][2])/DENOM;
	if(*r<0) *r = 0;
	else if(*r>255) *r = 255;
	if(*g<0) *g = 0;
	else if(*g>255) *g = 255;
	if(*b<0) *b = 0;
	else if(*b>255) *b = 255;
}

void Hue::mat2lmat(float mat[4][4], int lmat[4][4])
{
	int x, y;

	for (y = 0; y < 4; y++)
		for (x = 0; x < 4; x++) 
			lmat[y][x] = (int) (mat[y][x] * DENOM);
}


AVSValue __cdecl Create_Hue(AVSValue args, void* user_data, IScriptEnvironment* env) {
	return new Hue(args[0].AsClip(),
		args[1].AsFloat(0.0),
		args[2].AsFloat(1.0),
		args[3].AsFloat(1.0),
		args[4].AsString("RYGCBM"), env);
}

extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env)
{
  env->AddFunction("Hue", "c[hue]f[sat]f[intensity]f[channels]s", Create_Hue, 0);
  return 0;
}
